const T = 90
function experiment(rvec::Vector, exp_size::Number, steady::SteadyState, model::Model; PE = true, doMPC = true, incomeMPC = false, a0 = 0., retVF2 = false)
  #Perform transition
  # Starting backwards from the final steady state, iterate backward with the Bellman equation
  # Timing convention: Period 1 is the steady state.
  # Period 2 the shock hits. Still the steady state distribution, but policy and aggregates change
  finalVF = steady.VF
  initial_dist = steady.dist
  finalpolicy = steady.policy
  dim = length(size(finalVF))
  VFs = Array(Array{Float64, dim}, T) # will fill later
  dists = Array(Array{Float64, dim}, T)
  policys = Array(typeof(steady.policy), T); fill!(policys, steady.policy) # initialize
  bigpolicys = Array(typeof(steady.bigpolicy), T); fill!(bigpolicys, steady.bigpolicy)
  VFs[1], VFs[T] = copy(steady.VF), copy(steady.VF)
  dists[1], dists[2], dists[T] = copy(steady.dist), copy(steady.dist), copy(steady.dist)

  if !PE
    for t = T:-1:3
      VFs[t - 1] = copy(VFs[t])
      policys[t-1] = deepcopy(policys[t])
      VF = VFs[t - 1]
      policy = policys[t - 1]
      r = rvec[t - 1]
      if t == 3
        aggstate = makeAggState(r, rvec[3], a0, steady, model)
      else
        w = r2wage(r, model)
        aggstate = AggState(r, w, model)
      end
      bellman_operator2!(VF, policy, model, aggstate, SKIP_HOWARD = true)
    end
  else
    nothing
  end
  if retVF2
    return VFs[2]
  end

  newdist = zeros(dists[T])
  if doMPC && !incomeMPC
    dists[2] = MPC(exp_size, steady, model)
  end
  if incomeMPC
    dists[2] = doincomeMPC(exp_size, steady, model)
  end
  # timing: period 2 shock occurs
  # distribution is shocked, and policy functions change
  # period 1 is steady state
  aggss = Array(Aggregates, T)
  aggss[1] = steady.aggs
  aggss[T] = steady.aggs
  for t = 2:T-1
    policy = policys[t]
    bigpolicy = bigpolicys[t]
    makebigpolicy!(bigpolicy, policy, model)
    policy_mwork = 2 - (bigpolicy.gh_m .> 0)
    policy_fwork = 2 - (bigpolicy.gh_f .> 0)
    dists[t+1] = copy(dists[t])
    dist = dists[t+1]
    aggs = compute_aggregates(dist, model, bigpolicy) # First compute time t aggregates
    aggss[t] = aggs
    distribution_operator!(dist, bigpolicy.g, policy_mwork, policy_fwork, model, newdist) # Then update the t+1 distribution
  end
  #table2, table3 = makeTable2(steady, dists[2], model)
  #println(table2)
  #println(table3)
  return aggss
end

function MPC(exp_size, steady::SteadyState, model)
  ## a' = α + βa
  ## α s.t. Ea' = Ea, so α = (1 - β)Ea.
  dist = steady.dist
  newdist = zeros(steady.dist)
  n_a, n_s, n_ret, n_emp_m, n_emp_f, n_d = size(dist)
  MPCpolicy_nx = getMPCpolicy_nx(exp_size, steady, model)
  for s_nx in 1:n_s, ret_nx in 1:n_ret, emp_m_nx in 1:n_emp_m, emp_f_nx in 1:n_emp_f, d_nx in 1:n_d, a_nx in 1:n_a
    currentmass = dist[a_nx, s_nx, ret_nx, emp_m_nx, emp_f_nx, d_nx]
    currentmass == 0 && continue # Save time
    ap_nx = MPCpolicy_nx[a_nx, s_nx, ret_nx, emp_m_nx, emp_f_nx, d_nx]
    newdist[ap_nx, s_nx, ret_nx, emp_m_nx, emp_f_nx, d_nx] += currentmass
  end
  return newdist
end

function getMPCpolicy_nx(exp_size, steady::SteadyState, model)
  dist = steady.dist
  newdist = zeros(steady.dist)
  n_a, n_s, n_ret, n_emp_m, n_emp_f, n_d = size(dist)
  Ea = steady.aggs.A
  β = exp_size
  α = (1 - β)*Ea
  a_grid = model.biga_grid
  MPCpolicy = copy(steady.bigpolicy.gval)
  MPCpolicy_nx = copy(steady.bigpolicy.g)
  for s_nx in 1:n_s, ret_nx in 1:n_ret, emp_m_nx in 1:n_emp_m, emp_f_nx in 1:n_emp_f, d_nx in 1:n_d
    MPCpolicy[:, s_nx, ret_nx, emp_m_nx, emp_f_nx, d_nx] = α + β*a_grid
    MPCpolicy_nx[:, s_nx, ret_nx, emp_m_nx, emp_f_nx, d_nx] = locateslow2(slice(MPCpolicy, :, s_nx, ret_nx, emp_m_nx, emp_f_nx, d_nx ), a_grid)
  end
  return MPCpolicy_nx
end

function doincomeMPC(exp_size, steady::SteadyState, model)
  ## a' = α + βa
  ## α s.t. Ea' = Ea, so α = (1 - β)Ea.
  dist = steady.dist
  newdist = zeros(steady.dist)
  n_a, n_s, n_ret, n_emp_m, n_emp_f, n_d = size(dist)
  MPCpolicy_nx = getMPCpolicy_income_nx(exp_size, steady, model)
  for s_nx in 1:n_s, ret_nx in 1:n_ret, emp_m_nx in 1:n_emp_m, emp_f_nx in 1:n_emp_f, d_nx in 1:n_d, a_nx in 1:n_a
    currentmass = dist[a_nx, s_nx, ret_nx, emp_m_nx, emp_f_nx, d_nx]
    currentmass == 0 && continue # Save time
    ap_nx = MPCpolicy_nx[a_nx, s_nx, ret_nx, emp_m_nx, emp_f_nx, d_nx]
    newdist[ap_nx, s_nx, ret_nx, emp_m_nx, emp_f_nx, d_nx] += currentmass
  end
  return newdist
end

function getMPCpolicy_income_nx(exp_size, steady::SteadyState, model)
  dist = steady.dist
  newdist = zeros(steady.dist)
  policy = steady.bigpolicy
  g_income = policy.g_income
  n_a, n_s, n_ret, n_emp_m, n_emp_f, n_d = size(dist)
  Ey = steady.aggs.Y
  Ea = steady.aggs.A
  β = exp_size
  β = 1 - (1 - β)*Ea/Ey
  @show Ea/Ey
  @show β
  α = (1 - β)*Ey
  a_grid = model.biga_grid
  MPCpolicy = copy(steady.bigpolicy.gval)
  MPCpolicy_nx = copy(steady.bigpolicy.g)
  for s_nx in 1:n_s, ret_nx in 1:n_ret, emp_m_nx in 1:n_emp_m, emp_f_nx in 1:n_emp_f, d_nx in 1:n_d
    MPCpolicy[:, s_nx, ret_nx, emp_m_nx, emp_f_nx, d_nx] = a_grid + α - (1 - β)*g_income[:, s_nx, ret_nx, emp_m_nx, emp_f_nx, d_nx]
    MPCpolicy_nx[:, s_nx, ret_nx, emp_m_nx, emp_f_nx, d_nx] = locateslow2(slice(MPCpolicy, :, s_nx, ret_nx, emp_m_nx, emp_f_nx, d_nx ), a_grid)
  end
  return MPCpolicy_nx
end

function plotaggs(aggss::Vector{Aggregates}; plotnames::Vector{Symbol} = fieldnames(Aggregates), T = 15, leg2 = "", printout = false)
  numv = length(plotnames)
  leg = Array(AbstractString, numv)
  output = zeros(numv, T)
  figure()
  for vars_nx in 1:numv
    sym = plotnames[vars_nx]
    leg[vars_nx] = string(sym)
    ssval = getfield(aggss[1], sym)
    for t in 1:T
      output[vars_nx, t] = 100(getfield(aggss[t], sym)/ssval - 1)
      name = string(sym)
    end
    plot(2:T, slice(output, vars_nx, 2:T))
  end
  if leg2 != ""
    leg = leg2
  end
  legend(leg)
  if printout
    println(output[:, 2])
  end
end

function getdelA(aggss::Vector{Aggregates})
  ssval = getfield(aggss[1], :A)
  100(getfield(aggss[3], :A)/ssval - 1)
end

function GEexperiment(exp_size::Number, steady::SteadyState, model::Model; chi = 0., filename = "", rvec = steady.r*ones(T), retVF2 = false, incomeMPC = false)
  aggss = experiment(rvec, exp_size, steady, model, PE = false)
  rstarvec = Float64[rstar_GE(aggs, model) for aggs in aggss]
  if chi == 1.
    deltar = 0.999
  else
    deltar = 0.8
  end
  iter = 0
  while maximum(abs(rvec - rstarvec)) > 1e-5
    err = maximum(abs(rvec - rstarvec))
    if chi == 1.
      if err > 1e-2
        deltar = 0.99
      elseif 1e-3 < err <= 1e-2
        deltar = 0.99
      else
        deltar = 0.996
      end
    end
    rvec = (1 - deltar)*rstarvec + deltar*rvec
    aggss = experiment(rvec, exp_size, steady, model, PE = false, incomeMPC = incomeMPC)
    rstarvec = Float64[rstar_GE(aggs, model) for aggs in aggss]
    rstarvec = (rstarvec - Float64[chi*(aggs.A - steady.aggs.A) for aggs in aggss])
    rstarvec[2] = rstar_GE(aggss[2], model) # Correct for slight numerical error throwing off sensitive interest rates
    @show maximum(abs(rvec - rstarvec))
    if filename != ""
      res = (rvec, aggss, maximum(abs(rvec - rstarvec)), rstarvec)
      save("$filename.jld", "res", res)
    end
    iter += 1
  end
  @show "Hello!"
  if retVF2
    VF2 = experiment(rvec, exp_size, steady, model, PE = false, retVF2 = true, incomeMPC = incomeMPC)
    return VF2, aggss
  else
    return rvec, aggss
  end
end

function makeAggState{UF<:UtilityFunction}(r, r2, a0, steady, model::Model{UF, AggState})
  AggState(r, r2wage(r, model))
end

function makeAggState{UF<:UtilityFunction}(r, r2, a0, steady, model::Model{UF, AggStateSticky})
  r0 = steady.r
  rho = model.storage.etc[:rho]
  lambda = model.storage.etc[:lambda]
  Q1 = (1+r2)/(1+r0)
  w0 = r2wage(r0, model)
  w = r2wage(r, model)
  p1 = w
  w = ((p1^(1-rho)-lambda*(w0/Q1)^(1-rho))/(1-lambda))^(1/(1-rho))

  @show w w0
  AggStateSticky(r, w, a0)
end

function astar_GE(r, r2, aggs, steady, model)
  rho = model.storage.etc[:rho]
  lambda = model.storage.etc[:lambda]
  r0 = steady.r
  w0 = r2wage(r0, model)
  Q1 = (1+r2)/(1+r0)
  w = r2wage(r, model)
  p1 = w
  w = ((p1^(1-rho)-lambda*(w0/Q1)^(1-rho))/(1-lambda))^(1/(1-rho))
  L = aggs.L
  Hfix1 = L/(lambda+(1-lambda)*(w0/(w*Q1))^rho)
  astar = (w0/Q1-w)*lambda*Hfix1
end


function stickyGEexperiment(exp_size::Number, steady::SteadyState, model::Model; chi = 0., filename = "", rvec = steady.r*ones(T))
  iter = 0
  a = 0.
  if chi == 1.
    deltar = 0.999
  else
    deltar = 0.8
  end
  aggss = experiment(rvec, exp_size, steady, model, PE = false, a0 = a)
  astar = astar_GE(rvec[2], rvec[3], aggss[2], steady, model) # next period r
  rstarvec = Float64[rstar_GE(aggs, model) for aggs in aggss]
  while (maximum(abs(rvec - rstarvec)) > 1e-5 || abs(a - astar) > 1e-4 )
    err = maximum(abs(rvec - rstarvec))
    if chi == 1.
      if err > 1e-2
        deltar = 0.99
      elseif 1e-3 < err <= 1e-2
        deltar = 0.99
      else
        deltar = 0.996
      end
    end
    rvec = (1 - deltar)*rstarvec + deltar*rvec
    a = (1 - .9)*astar + .9*a
    @show a
    aggss = experiment(rvec, exp_size, steady, model, PE = false, a0 = a)
    rstarvec = Float64[rstar_GE(aggs, model) for aggs in aggss]
    rstarvec = (rstarvec - Float64[chi*(aggs.A - steady.aggs.A) for aggs in aggss])
    rstarvec[2] = rstar_GE(aggss[2], model) # Correct for slight numerical error throwing off sensitive interest rates
    astar = astar_GE(rvec[2], rvec[3], aggss[2], steady, model) # next period r
    @show maximum(abs(rvec - rstarvec))
    if filename != ""
      res = (rvec, aggss, maximum(abs(rvec - rstarvec)), rstarvec )
      save("$filename.jld", "res", res)
    end
    iter += 1
  end
  return rvec, a, aggss
end

function makeTable2(steady::SteadyState, newdist, model::Model)
  # Compute exact table 2
  # This computes aggregates, but gives the contribution by quintile subgroup.
  dist = steady.dist
  agrid = model.biga_grid
  n_a, n_s, n_ret, n_emp_m, n_emp_f, n_d = size(dist)
  # Grab the quintiles
  quint = [0.; 0.2; 0.4; 0.6; 0.8; 1.; .9; .95; .99]
  mu_male = model.IP.mu_male
  mu_female = model.IP.mu_female
  # Fix this so you use the heterogeneity in other types
  # Right now >20% of people can be at 0 assets
  ###### Asset quintile setup
  APDF = Float64[sum(dist[i, :]) for i in 1:length(agrid)]
  ACDF = cumsum(APDF)
  aquint = Int64[quantileCDF(ACDF, quint[i] ) for i in 1:9 ]
  aquint[1] = 0
  aquint[6] = length(ACDF)
  quintIND = UnitRange{Int64}[(aquint[i] + 1):aquint[i+1]  for i in 1:8]
  quintIND[6] = (aquint[7]+1):aquint[8]
  quintIND[7] = (aquint[8]+1):aquint[9]
  quintIND[8] = (aquint[9]+1):aquint[6]
  @show quintIND
  aggs = steady.aggs
  C, L, H, A = aggs.C, aggs.L, aggs.H, aggs.A
  policy = steady.bigpolicy
  g_c, gval, gh_m, gh_f = policy.g_c, policy.gval, policy.gh_m, policy.gh_f
  gh = (mu_female*gh_f + mu_male*gh_m)
  #gh = mu_male*gh_m
  gl = zeros(gh)
  s_vals = model.IP.s_vals
  for s_nx in 1:n_s, ret_nx in 1:n_ret, emp_m_nx in 1:n_emp_m, emp_f_nx in 1:n_emp_f, d_nx in 1:n_d, a_nx in 1:n_a
    IND = (a_nx, s_nx, ret_nx, emp_m_nx, emp_f_nx, d_nx)
    gl[IND...] = s_vals[s_nx, d_nx] * gh[IND...]
  end
  table2 = zeros(length(quintIND) + 1, 4)
  for q_nx in 1:(length(quintIND) + 1)
    if q_nx == length(quintIND) + 1
      IND = (1:(n_a), :, :, :, :, :)
      #IND = (:, :, :, :, :, :)
    else
      qIND = quintIND[q_nx]
      IND = (qIND, :, :, :, :, :)
    end
    table2[q_nx, 1] = 100sum(gval[IND...].*(newdist[IND...] - dist[IND...]))/A
    table2[q_nx, 2] = 100sum(g_c[IND...].*(newdist[IND...] - dist[IND...]))/C
    table2[q_nx, 3] = 100sum(gh[IND...].*(newdist[IND...] - dist[IND...]))/H
    table2[q_nx, 4] = 100sum(gl[IND...].*(newdist[IND...] - dist[IND...]))/L
    if q_nx == length(quintIND) + 1
      oldc = sum(g_c[IND...] .* dist[IND...])
      oldc2 = C
      @show oldc
      @show oldc2
    end
  end
  Lm = aggs.Lm
  Hm = aggs.Hm
  table3 = zeros(n_s + 2, 4)
  for q_nx in 1:n_s + 2
    if q_nx == n_s + 2
      IND = (1:(n_a), :, :, :, :, :)
      #IND = (:, :, :, :, :, :)
    elseif q_nx == n_s + 1
      IND = (1:(n_a), :, 2, :, :, :)
    else
      IND = (1:(n_a), q_nx, 1, :, :, :)
    end
    table3[q_nx, 1] = 100sum(gval[IND...].*(newdist[IND...] - dist[IND...]))/A
    table3[q_nx, 2] = 100sum(g_c[IND...].*(newdist[IND...] - dist[IND...]))/C
    table3[q_nx, 3] = 100sum(gh[IND...].*(newdist[IND...] - dist[IND...]))/H
    table3[q_nx, 4] = 100sum(gl[IND...].*(newdist[IND...] - dist[IND...]))/L
  end
  return table2, table3
end
