macro forwardrule(x, e)
  x, e = esc(x), esc(e)
  quote
    $e = sqrt(eps(eltype($x))) * max(one(eltype($x)), abs($x))
  end
end

macro hessianrule(x, e)
  x, e = esc(x), esc(e)
  quote
    $e = eps(eltype($x))^(1/4) * max(one(eltype($x)), abs($x))
  end
end

function forward_difference(f, x::Number)
  @forwardrule x h
  (f(x + h) - f(x)) / h
end

function second_difference(f, x::Number)
  @hessianrule x h
  (f(x+2h) - 2f(x+h) +f(x) )/h^2
end

function newton_opt(f, x::Float64)
  oldx = Inf
  for n = 1:100
    oldx = x
    fpx, fppx = forward_difference(f,x), second_difference(f,x)
    x = .5x + .5*(x - fpx/fppx) # convex combination
    abs(oldx - x) < 1e-6 && return x
  end
  println("Too many iterations")
  return NaN
end

function brent{T <: Float64}(f, x_lower::T, x_upper::T, x_guess::T = x_lower + 0.5 * (3.0 - sqrt(5.0))*(x_upper-x_lower) ;
                                   rel_tol::T = sqrt(eps(T)),
                                   abs_tol::T = eps(T),
                                   iterations::Integer = 1_000)

  if !(x_lower < x_upper)
    error("x_lower must be less than x_upper")
  end

  # Save for later
  initial_lower = x_lower
  initial_upper = x_upper

  const golden_ratio::T = 0.5 * (3.0 - sqrt(5.0))

  if(x_lower > x_guess || x_upper < x_guess)
    error("Guess out of bounds")
  end

  x_minimum = x_guess
  f_minimum = f(x_minimum)
  f_calls = 1 # Number of calls to f

  step = zero(T)
  step_old = zero(T)

  x_minimum_old = x_minimum
  x_minimum_old_old = x_minimum

  f_minimum_old = f_minimum
  f_minimum_old_old = f_minimum

  it = 0
  converged = false

  while it < iterations

    p = zero(T)
    q = zero(T)

    tolx = rel_tol * abs(x_minimum) + abs_tol

    x_midpoint = (x_upper+x_lower)/2

    if abs(x_minimum - x_midpoint) <= 2*tolx - (x_upper-x_lower)/2
      converged = true
      break
    end

    it += 1

    if abs(step_old) > tolx
      # Compute parabola interpolation
      # x_minimum + p/q is the optimum of the parabola
      # Also, q is guaranteed to be positive

      r = (x_minimum - x_minimum_old) * (f_minimum - f_minimum_old_old)
      q = (x_minimum - x_minimum_old_old) * (f_minimum - f_minimum_old)
      p = (x_minimum - x_minimum_old_old) * q - (x_minimum - x_minimum_old) * r
      q = 2(q - r)

      if q > 0
        p = -p
      else
        q = -q
      end
    end

    if abs(p) < abs(q*step_old/2) && p < q*(x_upper-x_minimum) && p < q*(x_minimum-x_lower)
      step_old = step
      step = p/q

      # The function must not be evaluated too close to x_upper or x_lower
      x_temp = x_minimum + step
      if ((x_temp - x_lower) < 2*tolx || (x_upper - x_temp) < 2*tolx)
        step = (x_minimum < x_midpoint) ? tolx : -tolx
      end
    else
      step_old = (x_minimum < x_midpoint) ? x_upper - x_minimum : x_lower - x_minimum
      step = golden_ratio * step_old
    end

    # The function must not be evaluated too close to x_minimum
    if abs(step) >= tolx
      x_new = x_minimum + step
    else
      x_new = x_minimum + ((step > 0) ? tolx : -tolx)
    end

    f_new = f(x_new)
    f_calls += 1

    if f_new <= f_minimum
      if x_new < x_minimum
        x_upper = x_minimum
      else
        x_lower = x_minimum
      end
      x_minimum_old_old = x_minimum_old
      f_minimum_old_old = f_minimum_old
      x_minimum_old = x_minimum
      f_minimum_old = f_minimum
      x_minimum = x_new
      f_minimum = f_new
    else
      if x_new < x_minimum
        x_lower = x_new
      else
        x_upper = x_new
      end
      if x_new <= x_minimum_old || x_minimum_old == x_minimum
        x_minimum_old_old = x_minimum_old
        f_minimum_old_old = f_minimum_old
        x_minimum_old = x_new
        f_minimum_old = f_new
      elseif x_new <= x_minimum_old_old || x_minimum_old_old == x_minimum || x_minimum_old_old == x_minimum_old
        x_minimum_old_old = x_new
        f_minimum_old_old = f_new
      end
    end
  end
  it == iterations && error("Brent's method failed to converge")
  return (x_minimum, f_minimum)
end


function myNewton(f, j, x::Number, lb = -Inf, ub = Inf)
  lcounter = 0
  ucounter = 0
  for n = 1:25
    fx , jx = f(x), j(x)
    abs(fx) < 1e-8 && return x
    d = fx/jx
    propose = x - d
    if lb < propose < ub
      x = propose
    elseif propose < lb
      x = (x + lb)/2
      lcounter += 1
    else
      x = (x + ub)/2
      ucounter += 1
    end
  end
  #fx = f(x)
  #println("Giving up on Newton method x = $x lb = $lb ub = $ub,
  #lc = $lcounter uc = $ucounter fx = $fx" )
  Roots.find_zero(f, lb, ub)
end


function myNewton2(f,x::Number, lb = -Inf, ub = Inf)
  if !(lb < x < ub)
    x = (lb + ub)/2
  end
  fx, jx = 0., 0.
  for n = 1:30
    jx, ar  = ForwardDiff.derivative(f, x, ForwardDiff.AllResults)
    fx = ForwardDiff.value(ar)
    abs(fx) < 1e-8 && return x
    d = fx/jx
    propose = x - d
    if lb < propose < ub
      x = propose
    elseif propose < lb
      x = (x + lb)/2
    else
      x = (x + ub)/2
    end
  end
  Roots.find_zero(f, lb, ub)
  #Roots.fzero(f, x, order = 0)
  #error("x = $x fx = $fx jx = $jx")
end

function my_opt(f, x::Number, lb, ub)
  j = ForwardDiff.derivative(f)
  myNewton2(j, x::Number, lb, ub)
end


function testprobit()
  # Model: Y* = XB + eps
  # Y = Y*>0
  # P(Y == 1) = P(XB + eps > 0)
  # = P(eps > -XB)
  mydist = Normal()
  beta = 0.5
  X = rand(100)
  Ys = beta*X + 0.5*randn(100)
  data = 1. * (Ys .> 0)

  function logL(betahat)
    logl = zero(betahat)
    for i = 1:length(X)
      if data[i] == 0.
        logl += log(cdf(mydist, -X[i]*betahat))
      else
        logl += log(1 - cdf(mydist, -X[i]*betahat))
      end
    end
    -logl
  end
optimize(logL, -10., 10.)

end
